//
//  LibytnefNestedHelper.m
//  TNEFsEnoughLib
//
//  Created by Josh Jacob on 3/28/14.
//  Copyright (c) 2014 Josh Jacob. All rights reserved.
//

#import "LibytnefNestedHelper.h"

#import "tnef-types.h"
#import "mapi.h"
#import "ytnef.h"

#define PRODID @"PRODID:-//TNEF's Enough//TNEF's Enough//EN\n"

//
// straight C functions
//

unsigned char GetRruleCount(unsigned char a, unsigned char b);
char * GetRruleDayname(unsigned char a);
unsigned char GetRruleMonthNum(unsigned char a, unsigned char b);

@implementation LibytnefNestedHelper

+ (NSString *)stringFromMapiProp:(MAPIProps)mapiProps propid:(DWORD)PROPID
{
    variableLength *vl;
    
    if ((vl = MAPIFindProperty(&mapiProps, PROP_TAG(PT_STRING8, PROPID))) != MAPI_UNDEFINED) {
        return [NSString stringWithCString:vl->data encoding:NSWindowsCP1252StringEncoding];
    } else if ((vl = MAPIFindProperty(&mapiProps, PROP_TAG(PT_UNICODE, PROPID))) != MAPI_UNDEFINED) {
        return [NSString stringWithUTF8String:vl->data];
    }
    
    return nil;
}

+ (NSString *)stringFromMapiUserProp:(MAPIProps)mapiProps propid:(DWORD)PROPID
{
    variableLength *vl;
    
    if ((vl = MAPIFindUserProp(&mapiProps, PROP_TAG(PT_STRING8, PROPID))) != MAPI_UNDEFINED) {
        return [NSString stringWithCString:vl->data encoding:NSWindowsCP1252StringEncoding];
    } else if ((vl = MAPIFindUserProp(&mapiProps, PROP_TAG(PT_UNICODE, PROPID))) != MAPI_UNDEFINED) {
        return [NSString stringWithUTF8String:vl->data];
    }
    
    return nil;
}

+ (NSString *)escapeString:(NSString *)string
{
    NSMutableString *mutantString = [[NSMutableString alloc] initWithString:string];
    
    [mutantString replaceOccurrencesOfString:@"\\" withString:@"\\\\" options:NSLiteralSearch range:NSMakeRange(0, [mutantString length])];
    [mutantString replaceOccurrencesOfString:@"\n" withString:@"\\n" options:NSLiteralSearch range:NSMakeRange(0, [mutantString length])];
    [mutantString replaceOccurrencesOfString:@"\r" withString:@"\\r" options:NSLiteralSearch range:NSMakeRange(0, [mutantString length])];
    [mutantString replaceOccurrencesOfString:@";" withString:@"\\;" options:NSLiteralSearch range:NSMakeRange(0, [mutantString length])];
    [mutantString replaceOccurrencesOfString:@"," withString:@"\\," options:NSLiteralSearch range:NSMakeRange(0, [mutantString length])];
    
    return mutantString;
}

+ (NSString *)variableLengthToString:(variableLength)varLength
{
    char *tempString = malloc(varLength.size + 1);
    strncpy(tempString, varLength.data, varLength.size);
    NSString *string = [NSString stringWithCString:tempString encoding:NSWindowsCP1252StringEncoding];
    free(tempString);
    return string;
}

+ (NSString *)PrintRTF:(variableLength *)VL
{
    NSMutableString *string = [[NSMutableString alloc] init];
    
    int index;
    char *byte;
    int brace_ct;
    int key;
    
    key = 0;
    brace_ct = 0;
    
    for(index = 0, byte=VL->data; index < VL->size; index++, byte++) {
        if (*byte == '}') {
            brace_ct--;
            key = 0;
            continue;
        }
        if (*byte == '{') {
            brace_ct++;
            continue;
        }
        if (*byte == '\\') {
            key = 1;
        }
        if (isspace(*byte)) {
            key = 0;
        }
        if ((brace_ct == 1) && (key == 0)) {
            if (*byte == '\n') {
                [string appendString:@"\\n"];
            } else if (*byte == '\r') {
                // Print nothing.
            } else if (*byte == ';') {
                [string appendString:@"\\;"];
            } else if (*byte == ',') {
                [string appendString:@"\\,"];
            } else if (*byte == '\\') {
                [string appendString:@"\\"];
            } else {
                [string appendFormat:@"%c", *byte];
            }
        }
    }
    [string appendString:@"\n"];
    return string;
}

unsigned char GetRruleCount(unsigned char a, unsigned char b)
{
    return ((a << 8) | b);
}

unsigned char GetRruleMonthNum(unsigned char a, unsigned char b)
{
    switch (a) {
        case 0x00:
            switch (b) {
                case 0x00:
                    // Jan
                    return(1);
                case 0xA3:
                    // May
                    return(5);
                case 0xAE:
                    // Nov
                    return(11);
            }
            break;
        case 0x60:
            switch (b) {
                case 0xAE:
                    // Feb
                    return(2);
                case 0x51:
                    // Jun
                    return(6);
            }
            break;
        case 0xE0:
            switch (b) {
                case 0x4B:
                    // Mar
                    return(3);
                case 0x56:
                    // Sep
                    return(9);
            }
            break;
        case 0x40:
            switch (b) {
                case 0xFA:
                    // Apr
                    return(4);
            }
            break;
        case 0x20:
            if (b == 0xFA) {
                // Jul
                return(7);
            }
            break;
        case 0x80:
            if (b == 0xA8) {
                // Aug
                return(8);
            }
            break;
        case 0xA0:
            if (b == 0xFF) {
                // Oct
                return(10);
            }
            break;
        case 0xC0:
            if (b == 0x56) {
                return(12);
            }
    }
    
    // Error
    return(0);
}

char * GetRruleDayname(unsigned char a)
{
    static char daystring[25];
    
    *daystring = 0;
    
    if (a & 0x01) {
        strcat(daystring, "SU,");
    }
    if (a & 0x02) {
        strcat(daystring, "MO,");
    }
    if (a & 0x04) {
        strcat(daystring, "TU,");
    }
    if (a & 0x08) {
        strcat(daystring, "WE,");
    }
    if (a & 0x10) {
        strcat(daystring, "TH,");
    }
    if (a & 0x20) {
        strcat(daystring, "FR,");
    }
    if (a & 0x40) {
        strcat(daystring, "SA,");
    }
    
    if (strlen(daystring)) {
        daystring[strlen(daystring) - 1] = 0;
    }
    
    return(daystring);
}

+ (NSString *)PrintRrule:(char *)recur_data size:(int) size tnef:(TNEFStruct)TNEF
{
    NSMutableString *string = [[NSMutableString alloc] init];
    
    variableLength *filename;
    
    if (size < 0x1F) {
        return nil;
    }
    
    [string appendString:@"RRULE:FREQ="];
    
    if (recur_data[0x04] == 0x0A) {
        [string appendString:@"DAILY"];
        
        if (recur_data[0x16] == 0x23 || recur_data[0x16] == 0x22 ||
            recur_data[0x16] == 0x21) {
            if ((filename=MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_I2, 0x0011))) != MAPI_UNDEFINED) {
               [string appendFormat:@";INTERVAL=%d", *(filename->data)];
            }
            if (recur_data[0x16] == 0x22 || recur_data[0x16] == 0x21) {
                [string appendFormat:@";COUNT=%d", GetRruleCount(recur_data[0x1B], recur_data[0x1A])];
            }
        } else if (recur_data[0x16] == 0x3E) {
            [string appendString:@";BYDAY=MO,TU,WE,TH,FR"];
            if (recur_data[0x1A] == 0x22 || recur_data[0x1A] == 0x21) {
                [string appendFormat:@";COUNT=%d", GetRruleCount(recur_data[0x1F], recur_data[0x1E])];
            }
        }
    } else if (recur_data[0x04] == 0x0B) {
        [string appendFormat:@"WEEKLY;INTERVAL=%d;BYDAY=%s", recur_data[0x0E], GetRruleDayname(recur_data[0x16])];
        if (recur_data[0x1A] == 0x22 || recur_data[0x1A] == 0x21) {
            [string appendFormat:@";COUNT=%d", GetRruleCount(recur_data[0x1F], recur_data[0x1E])];
        }
    } else if (recur_data[0x04] == 0x0C) {
        [string appendString:@"MONTHLY"];
        if (recur_data[0x06] == 0x02) {
            [string appendFormat:@";INTERVAL=%d;BYMONTHDAY=%d", recur_data[0x0E], recur_data[0x16]];
            if (recur_data[0x1A] == 0x22 || recur_data[0x1A] == 0x21) {
                [string appendFormat:@";COUNT=%d", GetRruleCount(recur_data[0x1F], recur_data[0x1E])];
            }
        } else if (recur_data[0x06] == 0x03) {
            [string appendFormat:@";BYDAY=%s;BYSETPOS=%d;INTERVAL=%d",
                    GetRruleDayname(recur_data[0x16]),
                    recur_data[0x1A] == 0x05 ? -1 : recur_data[0x1A],
                    recur_data[0x0E]];
            if (recur_data[0x1E] == 0x22 || recur_data[0x1E] == 0x21) {
                [string appendFormat:@";COUNT=%d", GetRruleCount(recur_data[0x23], recur_data[0x22])];
            }
        }
    } else if (recur_data[0x04] == 0x0D) {
        [string appendFormat:@"YEARLY;BYMONTH=%d", GetRruleMonthNum(recur_data[0x0A], recur_data[0x0B])];
        if (recur_data[0x06] == 0x02) {
            [string appendFormat:@";BYMONTHDAY=%d", recur_data[0x16]];
        } else if (recur_data[0x06] == 0x03) {
            [string appendFormat:@";BYDAY=%s;BYSETPOS=%d",
                    GetRruleDayname(recur_data[0x16]),
                    recur_data[0x1A] == 0x05 ? -1 : recur_data[0x1A]];
        }
        if (recur_data[0x1E] == 0x22 || recur_data[0x1E] == 0x21) {
            [string appendFormat:@";COUNT=%d", GetRruleCount(recur_data[0x23], recur_data[0x22])];
        }
    }
    
    [string appendString:@"\n"];
    return string;
}

+ (void)appendProperty:(NSMutableString *)string tnef:(TNEFStruct)TNEF proptype:(DWORD)PROPTYPE prodid:(DWORD)PROPID format:(NSString *)format
{
    variableLength *vl;
    if ((vl=MAPIFindProperty(&(TNEF.MapiProperties), PROP_TAG(PROPTYPE, PROPID))) != MAPI_UNDEFINED) {
        if (vl->size > 0) {
            if ((vl->size == 1) && (vl->data[0] == 0)) {
            } else {
                if (PROPTYPE == PT_STRING8) {
                    [string appendFormat:format, [NSString stringWithCString:vl->data encoding:NSWindowsCP1252StringEncoding]];
                } else if (PROPTYPE == PT_UNICODE) {
                    [string appendFormat:format, [NSString stringWithUTF8String:vl->data]];
                } else {
                    [string appendFormat:format, vl->data];
                }
            }
        }
    }
}

+ (void)appendStringProperty:(NSMutableString *)string tnef:(TNEFStruct)TNEF prodid:(DWORD)PROPID format:(NSString *)format
{
    NSString *stringProperty = [self stringFromMapiProp:TNEF.MapiProperties propid:PROPID];
    
    if (stringProperty != nil) {
        [string appendFormat:format, stringProperty];
    }
}

+ (void)appendUserProperty:(NSMutableString *)string tnef:(TNEFStruct)TNEF proptype:(DWORD)PROPTYPE prodid:(DWORD)PROPID format:(NSString *)format
{
    variableLength *vl;
    if ((vl=MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PROPTYPE, PROPID))) != MAPI_UNDEFINED) {
        if (vl->size > 0) {
            if ((vl->size == 1) && (vl->data[0] == 0)) {
            } else {
                if (PROPTYPE == PT_STRING8) {
                    [string appendFormat:format, [NSString stringWithCString:vl->data encoding:NSWindowsCP1252StringEncoding]];
                } else if (PROPTYPE == PT_UNICODE) {
                    [string appendFormat:format, [NSString stringWithUTF8String:vl->data]];
                } else {
                    [string appendFormat:format, vl->data];
                }
            }
        }
    }
}

+ (void)appendUserStringProperty:(NSMutableString *)string tnef:(TNEFStruct)TNEF prodid:(DWORD)PROPID format:(NSString *)format
{
    NSString *stringProperty = [self stringFromMapiProp:TNEF.MapiProperties propid:PROPID];
    
    if (stringProperty != nil) {
        [string appendFormat:format, stringProperty];
    }
}

+ (void)appendQuoted:(NSMutableString *)string variableLength:(variableLength *)VL
{
    int index;
    char buffer[2];
    buffer[1] = 0x00;
    
    for (index=0;index<VL->size-1; index++) {
        if (VL->data[index] == '\n') {
            [string appendString:@"=0A"];
        } else if (VL->data[index] == '\r') {
        } else {
            buffer[0] = VL->data[index];
            [string appendString:[NSString stringWithCString:buffer encoding:NSWindowsCP1252StringEncoding]];
        }
    }
}

+ (void)appendQuoted:(NSMutableString *)string string:(NSString *)appendingString
{
    // int index;
    char buffer[2];
    buffer[1] = 0x00;
    
    NSMutableString *escapedString = [[NSMutableString alloc] initWithString:appendingString];
    [escapedString replaceOccurrencesOfString:@"\n" withString:@"=0A" options:NSCaseInsensitiveSearch range:NSMakeRange(0, [escapedString length])];
    [escapedString replaceOccurrencesOfString:@"\r" withString:@"" options:NSCaseInsensitiveSearch range:NSMakeRange(0, [escapedString length])];
    
    [string appendString:escapedString];
}

//
// PUBLIC
//

+ (NSString *)SaveVTask:(TNEFStruct)TNEF
{
    NSMutableString *string = [[NSMutableString alloc] init];
    
    variableLength *filename;
    char *charptr, *charptr2;
    dtr thedate;
    DDWORD *ddword_ptr;
    DDWORD ddword_val;
    
    [string appendFormat:@"BEGIN:VCALENDAR\n"];
    [string appendString:PRODID];
    [string appendString:@"VERSION:2.0\n"];
    [string appendString:@"METHOD:PUBLISH\n" ];
    [string appendString:@"BEGIN:VTODO\n"];
    
    filename = NULL;
    
    if (TNEF.messageID[0] != 0) {
        [string appendFormat:@"UID:%@\n", [NSString stringWithCString:TNEF.messageID encoding:NSWindowsCP1252StringEncoding]];
    }
    filename = MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_STRING8, 0x8122));
    if (filename != MAPI_UNDEFINED) {
        NSString *tempString = [self variableLengthToString:*filename];
        [string appendFormat:@"ORGANIZER:%@\n", tempString];
    }
    
    if ((filename = MAPIFindProperty(&(TNEF.MapiProperties), PROP_TAG(PT_STRING8, PR_DISPLAY_TO))) != MAPI_UNDEFINED) {
        filename = MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_STRING8, 0x811f));
    }
    if ((filename != MAPI_UNDEFINED) && (filename->size > 1)) {
        charptr = filename->data-1;
        charptr2 = strstr(charptr+1, ";");
        while (charptr != NULL) {
            charptr++;
            charptr2 = strstr(charptr, ";");
            if (charptr2 != NULL) {
                *charptr2 = 0;
            }
            while (*charptr == ' ')
                charptr++;
            [string appendFormat:@"ATTENDEE;CN=%@;ROLE=REQ-PARTICIPANT:%@\n",
             [NSString stringWithCString:charptr encoding:NSWindowsCP1252StringEncoding],
             [NSString stringWithCString:charptr encoding:NSWindowsCP1252StringEncoding]];
            charptr = charptr2;
        }
    }
    
    if (TNEF.subject.size > 0) {
        [string appendString:@"SUMMARY:"];
        NSString *tempString = [self variableLengthToString:TNEF.subject];
        [string appendString:[self escapeString:tempString]];
        [string appendString:@"\n"];
    }
    
    if (TNEF.body.size > 0) {
        [string appendString:@"DESCRIPTION:"];
        NSString *tempString = [self variableLengthToString:TNEF.body];
        [string appendString:[self escapeString:tempString]];
        [string appendString:@"\n"];
    } else {
        filename = MAPIFindProperty(&(TNEF.MapiProperties), PROP_TAG(PT_UNICODE, PR_PREVIEW));
        if (filename != MAPI_UNDEFINED) {
            [string appendString:@"DESCRIPTION:"];
            NSString *tempString = [self variableLengthToString:*filename];
            [string appendString:[self escapeString:tempString]];
            [string appendString:@"\n"];
        }
    }
    
    filename = MAPIFindProperty(&(TNEF.MapiProperties), PROP_TAG(PT_SYSTIME, PR_CREATION_TIME));
    if (filename != MAPI_UNDEFINED) {
        [string appendString:@"DTSTAMP:"];
        MAPISysTimetoDTR(filename->data, &thedate);
        [string appendFormat:@"%04i%02i%02iT%02i%02i%02iZ\n",
                thedate.wYear, thedate.wMonth, thedate.wDay,
                thedate.wHour, thedate.wMinute, thedate.wSecond];
    }
    
    filename = MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_SYSTIME, 0x8517));
    if (filename != MAPI_UNDEFINED) {
        [string appendString:@"DUE:"];
        MAPISysTimetoDTR(filename->data, &thedate);
        [string appendFormat:@"%04i%02i%02iT%02i%02i%02iZ\n",
                thedate.wYear, thedate.wMonth, thedate.wDay,
                thedate.wHour, thedate.wMinute, thedate.wSecond];
    }
    filename = MAPIFindProperty(&(TNEF.MapiProperties), PROP_TAG(PT_SYSTIME, PR_LAST_MODIFICATION_TIME));
    if (filename != MAPI_UNDEFINED) {
        [string appendString:@"LAST-MODIFIED:"];
        MAPISysTimetoDTR(filename->data, &thedate);
        [string appendFormat:@"%04i%02i%02iT%02i%02i%02iZ\n",
                thedate.wYear, thedate.wMonth, thedate.wDay,
                thedate.wHour, thedate.wMinute, thedate.wSecond];
    }
    
    // Class
    filename = MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_BOOLEAN, 0x8506));
    if (filename != MAPI_UNDEFINED) {
        ddword_ptr = (DDWORD*)filename->data;
        ddword_val = SwapDDWord((BYTE*)ddword_ptr);
        [string appendString:@"CLASS:"];
        if (*ddword_ptr == 1) {
            [string appendString:@"PRIVATE\n"];
        } else {
            [string appendString:@"PUBLIC\n"];
        }
    }
    
    [string appendString:@"END:VTODO\n"];
    [string appendString:@"END:VCALENDAR\n"];
    
    return string;
}

+ (NSString *)SaveVCalendar:(TNEFStruct)TNEF
{
    NSMutableString *string = [[NSMutableString alloc] init];
    NSString *tempString = nil;

    variableLength *filename;
    char *charptr, *charptr2;
    int index;
    DDWORD *ddword_ptr;
    DDWORD ddword_val;
    dtr thedate;
    
    [string appendString:@"BEGIN:VCALENDAR\n"];
    if (TNEF.messageClass[0] != 0) {
        charptr2=TNEF.messageClass;
        charptr=charptr2;
        while (*charptr != 0) {
            if (*charptr == '.') {
                charptr2 = charptr;
            }
            charptr++;
        }
        if (strcmp(charptr2, ".MtgCncl") == 0) {
            [string appendString:@"METHOD:CANCEL\n"];
        } else {
            [string appendString:@"METHOD:REQUEST\n"];
        }
    } else {
        [string appendString:@"METHOD:REQUEST\n"];
    }
    [string appendString:PRODID];
    [string appendString:@"VERSION:2.0\n"];
    [string appendString:@"BEGIN:VEVENT\n"];
    
    // UID
    // After alot of comparisons, I'm reasonably sure this is totally
    // wrong.  But it's not really necessary.
    //
    // I think it only exists to connect future modification entries to
    // this entry. so as long as it's incorrectly interpreted the same way
    // every time, it should be ok :)
    filename = NULL;
    if ((filename=MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_BINARY, 0x3))) == MAPI_UNDEFINED) {
        if ((filename=MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_BINARY, 0x23))) == MAPI_UNDEFINED) {
            filename = NULL;
        }
    }
    if (filename!=NULL) {
        [string appendString:@"UID:"];
        for(index=0;index<filename->size;index++) {
            [string appendFormat:@"%02X", (unsigned char)filename->data[index]];
        }
        [string appendString:@"\n"];
    }
    
    // Sequence
    filename = NULL;
    if ((filename=MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_LONG, 0x8201))) != MAPI_UNDEFINED) {
        ddword_ptr = (DDWORD*)filename->data;
        [string appendFormat:@"SEQUENCE:%i\n", (int) *ddword_ptr];
    }
    
    // find sender
    filename = NULL;
    if ((filename=MAPIFindProperty(&(TNEF.MapiProperties), PROP_TAG(PT_BINARY, PR_SENDER_SEARCH_KEY))) == MAPI_UNDEFINED) {
        if ((filename=MAPIFindProperty(&(TNEF.MapiProperties), PROP_TAG(PT_STRING8, PR_SENDER_EMAIL_ADDRESS))) == MAPI_UNDEFINED) {
            if ((filename=MAPIFindProperty(&(TNEF.MapiProperties), PROP_TAG(PT_UNICODE, PR_SENDER_EMAIL_ADDRESS))) == MAPI_UNDEFINED) {
                filename = NULL;
            }
        }
    }
    if (filename!=NULL) {
        charptr = filename->data;
        charptr2 = strstr(charptr, ":");
        if (charptr2 == NULL)
            charptr2 = charptr;
        else
            charptr2++;
        [string appendFormat:@"ORGANIZER;CN=\"%s\":MAILTO:%s\n", charptr2, charptr2];
    }
    
    // Required Attendees
    if ((filename = MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_STRING8, 0x823b))) != MAPI_UNDEFINED) {
        // We have a list of required participants, so
        // write them out.
        if (filename->size > 1) {
            charptr = filename->data-1;
            charptr2=strstr(charptr+1, ";");
            while (charptr != NULL) {
                charptr++;
                charptr2 = strstr(charptr, ";");
                if (charptr2 != NULL) {
                    *charptr2 = 0;
                }
                while (*charptr == ' ')
                    charptr++;
                [string appendFormat:@"ATTENDEE;PARTSTAT=NEEDS-ACTION;"];
                [string appendFormat:@"ROLE=REQ-PARTICIPANT;RSVP=TRUE;"];
                [string appendFormat:@"CN=\"%s\":MAILTO:%s\n", charptr, charptr];
                charptr = charptr2;
            }
        }
        // Optional attendees
        if ((filename = MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_STRING8, 0x823c))) != MAPI_UNDEFINED) {
            // The list of optional participants
            if (filename->size > 1) {
                charptr = filename->data-1;
                charptr2=strstr(charptr+1, ";");
                while (charptr != NULL) {
                    charptr++;
                    charptr2 = strstr(charptr, ";");
                    if (charptr2 != NULL) {
                        *charptr2 = 0;
                    }
                    while (*charptr == ' ')
                        charptr++;
                    [string appendFormat:@"ATTENDEE;PARTSTAT=NEEDS-ACTION;"];
                    [string appendFormat:@"ROLE=OPT-PARTICIPANT;RSVP=TRUE;"];
                    [string appendFormat:@"CN=\"%s\":MAILTO:%s\n", charptr, charptr];
                    charptr = charptr2;
                }
            }
        }
    } else if ((filename = MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_STRING8, 0x8238))) != MAPI_UNDEFINED) {
        if (filename->size > 1) {
            charptr = filename->data-1;
            charptr2=strstr(charptr+1, ";");
            while (charptr != NULL) {
                charptr++;
                charptr2 = strstr(charptr, ";");
                if (charptr2 != NULL) {
                    *charptr2 = 0;
                }
                while (*charptr == ' ')
                    charptr++;
                [string appendFormat:@"ATTENDEE;PARTSTAT=NEEDS-ACTION;"];
                [string appendFormat:@"ROLE=REQ-PARTICIPANT;RSVP=TRUE;"];
                [string appendFormat:@"CN=\"%s\":MAILTO:%s\n", charptr, charptr];
                charptr = charptr2;
            }
        }
        
    }
    // Summary
    filename = NULL;
    if ((tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_CONVERSATION_TOPIC]) != nil) {
    // if((filename=MAPIFindProperty(&(TNEF.MapiProperties), PROP_TAG(PT_STRING8, PR_CONVERSATION_TOPIC))) != MAPI_UNDEFINED) {
        [string appendString:@"SUMMARY:"];
        // NSString *tempString = [self variableLengthToString:*filename];
        [string appendString:[self escapeString:tempString]];
        [string appendString:@"\n"];
    }
    
    // Description
    bool descriptionPrinted = false;
    if ((filename=MAPIFindProperty(&(TNEF.MapiProperties), PROP_TAG(PT_BINARY, PR_RTF_COMPRESSED))) != MAPI_UNDEFINED) {
        variableLength buf;
        if ((buf.data = DecompressRTF(filename, &(buf.size))) != NULL) {
            
            NSData *data = [[NSData alloc] initWithBytes:buf.data length:buf.size];
            NSAttributedString *body = nil;
            
#ifdef __MAC_OS_X_VERSION_MAX_ALLOWED
            body = [[NSAttributedString alloc] initWithRTF:data documentAttributes: nil];
#endif
#ifndef __MAC_OS_X_VERSION_MAX_ALLOWED
            body = [[NSData alloc] initWithData:data];
#endif
            if (body != nil) {
                [string appendString:@"DESCRIPTION;QUOTED-PRINTABLE:"];
                // [string appendString:[self PrintRTF:&buf]];
                [self appendQuoted:string string:[body string]];
                [string appendString:@"\n"];
                descriptionPrinted= true;
            }
            free(buf.data);
        }
    }
    if ((descriptionPrinted == false) && (TNEF.body.size > 0)) {
        [string appendString:@"DESCRIPTION:"];
        NSString *tempString = [self variableLengthToString:TNEF.body];
        [string appendString:[self escapeString:tempString]];
        [string appendString:@"\n"];
        descriptionPrinted= true;
    }
    if (descriptionPrinted == false) {
        tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_PREVIEW];
        if (tempString != nil) {
            [string appendString:@"DESCRIPTION:"];
            [string appendString:[self escapeString:tempString]];
            [string appendString:@"\n"];
        }
    }

    // Location
    filename = NULL;
    if ((tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:0x0002]) == nil) {
    // if ((filename=MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_STRING8, 0x0002))) == MAPI_UNDEFINED) {
        if ((tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:0x8208]) == nil) {
        // if ((filename=MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_STRING8, 0x8208))) == MAPI_UNDEFINED) {
            // filename = NULL;
            tempString = nil;
        }
    }
    if (tempString != nil) {
        [string appendString:@"LOCATION: "];
        [string appendString:tempString];
        [string appendString:@"\n"];
    }
    // Date Start
    filename = NULL;
    if ((filename=MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_SYSTIME, 0x820d))) == MAPI_UNDEFINED) {
        if ((filename=MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_SYSTIME, 0x8516))) == MAPI_UNDEFINED) {
            filename=NULL;
        }
    }
    if (filename != NULL) {
        [string appendString:@"DTSTART:"];
        MAPISysTimetoDTR(filename->data, &thedate);
        [string appendFormat:@"%04i%02i%02iT%02i%02i%02iZ\n",
                thedate.wYear, thedate.wMonth, thedate.wDay,
                thedate.wHour, thedate.wMinute, thedate.wSecond];
    }
    // Date End
    filename = NULL;
    if ((filename=MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_SYSTIME, 0x820e))) == MAPI_UNDEFINED) {
        if ((filename=MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_SYSTIME, 0x8517))) == MAPI_UNDEFINED) {
            filename=NULL;
        }
    }
    if (filename != NULL) {
        [string appendString:@"DTEND:"];
        MAPISysTimetoDTR(filename->data, &thedate);
        [string appendFormat:@"%04i%02i%02iT%02i%02i%02iZ\n",
                thedate.wYear, thedate.wMonth, thedate.wDay,
                thedate.wHour, thedate.wMinute, thedate.wSecond];
    }
    // Date Stamp
    filename = NULL;
    if ((filename=MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_SYSTIME, 0x8202))) != MAPI_UNDEFINED) {
        [string appendString:@"CREATED:"];
        MAPISysTimetoDTR(filename->data, &thedate);
        [string appendFormat:@"%04i%02i%02iT%02i%02i%02iZ\n",
                thedate.wYear, thedate.wMonth, thedate.wDay,
                thedate.wHour, thedate.wMinute, thedate.wSecond];
    }
    // Class
    filename = NULL;
    if ((filename=MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_BOOLEAN, 0x8506))) != MAPI_UNDEFINED) {
        ddword_ptr = (DDWORD*)filename->data;
        ddword_val = SwapDDWord((BYTE*)ddword_ptr);
        [string appendString:@"CLASS:"];
        if (*ddword_ptr == 1) {
            [string appendString:@"PRIVATE\n"];
        } else {
            [string appendString:@"PUBLIC\n"];
        }
    }
    // Recurrence
    filename = NULL;
    if ((filename=MAPIFindUserProp(&(TNEF.MapiProperties), PROP_TAG(PT_BINARY, 0x8216))) != MAPI_UNDEFINED) {
        [string appendString:[self PrintRrule:filename->data size:filename->size tnef:TNEF]];
    }
    
    // Wrap it up
    [string appendString:@"END:VEVENT\n"];
    [string appendString:@"END:VCALENDAR\n"];
    
    return string;
}


+ (NSString *)SaveVCard:(TNEFStruct)TNEF
{
    NSMutableString *string = [[NSMutableString alloc] init];
    
    variableLength *vl;
    NSString *pobox, *street, *city, *state, *zip, *country;
    dtr thedate;
    int boolean;
    
    NSString *tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_DISPLAY_NAME];
    if ((tempString == nil) || (tempString.length < 1)) {
        tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_COMPANY_NAME];
    }
    
    [string appendString:@"BEGIN:VCARD\n"];
    [string appendString:@"VERSION:2.1\n"];
    if (tempString != nil) {
        [string appendFormat:@"FN:%@\n", tempString];
        
        // N:Public;John;Quinlan;Mr.;Esq.
        [string appendString:@"N:"];
        [string appendFormat:@"%@;", (tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_SURNAME]) ? tempString : @""];
        [string appendFormat:@"%@;", (tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_GIVEN_NAME]) ? tempString : @""];
        [string appendFormat:@"%@;", (tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_MIDDLE_NAME]) ? tempString : @""];
        [string appendFormat:@"%@;", (tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_DISPLAY_NAME_PREFIX]) ? tempString : @""];
        [string appendFormat:@"%@", (tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_GENERATION]) ? tempString : @""];
        [string appendString:@"\n"];
    }
    
    [self appendStringProperty:string tnef:TNEF prodid:PR_NICKNAME format:@"NICKNAME:%@\n"];
    [self appendUserStringProperty:string tnef:TNEF prodid:0x8554 format:@"MAILER:Microsoft Outlook %@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_SPOUSE_NAME format:@"X-EVOLUTION-SPOUSE:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_MANAGER_NAME format:@"X-EVOLUTION-MANAGER:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_ASSISTANT format:@"X-EVOLUTION-ASSISTANT:%@\n"];
    
    // Organizational
    if ((tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_COMPANY_NAME]) != nil) {
        if (tempString.length > 0) {
            [string appendFormat:@"ORG:%@", tempString];
            if ((tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_DEPARTMENT_NAME]) != nil) {
                if (tempString.length > 0) {
                    [string appendFormat:@";%@", tempString];
                }
            }
            [string appendString:@"\n"];
        }
    }
    
    [self appendStringProperty:string tnef:TNEF prodid:PR_OFFICE_LOCATION format:@"X-EVOLUTION-OFFICE:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_TITLE format:@"TITLE:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_PROFESSION format:@"ROLE:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_BODY format:@"NOTE:%@\n"];
    if (TNEF.body.size > 0) {
        [string appendString:@"NOTE;QUOTED-PRINTABLE:"];
        [self appendQuoted:string variableLength:&(TNEF.body)];
        [string appendString:@"\n"];
    }
    
    
    // Business Address
    boolean = 0;
    if ((pobox = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_POST_OFFICE_BOX]) != nil) {
        boolean = 1;
    }
    if ((street = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_STREET_ADDRESS]) != nil) {
        boolean = 1;
    }
    if ((city = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_LOCALITY]) != nil) {
        boolean = 1;
    }
    if ((state = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_STATE_OR_PROVINCE]) != nil) {
        boolean = 1;
    }
    if ((zip = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_POSTAL_CODE]) != nil) {
        boolean = 1;
    }
    if ((country = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_COUNTRY]) != nil) {
        boolean = 1;
    }
    if (boolean == 1) {
        [string appendString:@"ADR;QUOTED-PRINTABLE;WORK:"];
        if (pobox != nil) {
            [self appendQuoted:string string:pobox];
        }
        [string appendString:@";;"];
        if (street != nil) {
            [self appendQuoted:string string:street];
        }
        [string appendString:@";"];
        if (city != nil) {
            [self appendQuoted:string string:city];
        }
        [string appendString:@";"];
        if (state != nil) {
            [self appendQuoted:string string:state];
        }
        [string appendString:@";"];
        if (zip != nil) {
            [self appendQuoted:string string:zip];
        }
        [string appendString:@";"];
        if (country != nil) {
            [self appendQuoted:string string:country];
        }
        [string appendString:@"\n"];
//        if ((tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:0x801b]) != nil) {
//            [string appendString:@"LABEL;QUOTED-PRINTABLE;WORK:"];
//            [string appendString:tempString];
//            [string appendString:@"\n"];
//        }
    }
    
    // Home Address
    boolean = 0;
    if ((pobox = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_HOME_ADDRESS_POST_OFFICE_BOX]) != nil) {
        boolean = 1;
    }
    if ((street = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_HOME_ADDRESS_STREET]) != nil) {
        boolean = 1;
    }
    if ((city = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_HOME_ADDRESS_CITY]) != nil) {
        boolean = 1;
    }
    if ((state = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_HOME_ADDRESS_STATE_OR_PROVINCE]) != nil) {
        boolean = 1;
    }
    if ((zip = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_HOME_ADDRESS_POSTAL_CODE]) != nil) {
        boolean = 1;
    }
    if ((country = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_HOME_ADDRESS_COUNTRY]) != nil) {
        boolean = 1;
    }
    if (boolean == 1) {
        [string appendString:@"ADR;QUOTED-PRINTABLE;HOME:"];
        if (pobox != nil) {
            [self appendQuoted:string string:pobox];
        }
        [string appendString:@";;"];
        if (street != nil) {
            [self appendQuoted:string string:street];
        }
        [string appendString:@";"];
        if (city != nil) {
            [self appendQuoted:string string:city];
        }
        [string appendString:@";"];
        if (state != nil) {
            [self appendQuoted:string string:state];
        }
        [string appendString:@";"];
        if (zip != nil) {
            [self appendQuoted:string string:zip];
        }
        [string appendString:@";"];
        if (country != nil) {
            [self appendQuoted:string string:country];
        }
        [string appendString:@"\n"];
//        if ((tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:0x801a]) != nil) {
//            [string appendString:@"LABEL;QUOTED-PRINTABLE;WORK:"];
//            [string appendString:tempString];
//            [string appendString:@"\n"];
//        }
    }
    
    // Other Address
    boolean = 0;
    if ((pobox = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_OTHER_ADDRESS_POST_OFFICE_BOX]) != nil) {
        boolean = 1;
    }
    if ((street = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_OTHER_ADDRESS_STREET]) != nil) {
        boolean = 1;
    }
    if ((city = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_OTHER_ADDRESS_CITY]) != nil) {
        boolean = 1;
    }
    if ((state = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_OTHER_ADDRESS_STATE_OR_PROVINCE]) != nil) {
        boolean = 1;
    }
    if ((zip = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_OTHER_ADDRESS_POSTAL_CODE]) != nil) {
        boolean = 1;
    }
    if ((country = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_OTHER_ADDRESS_COUNTRY]) != nil) {
        boolean = 1;
    }
    if (boolean == 1) {
        [string appendString:@"ADR;QUOTED-PRINTABLE;OTHER:"];
        if (pobox != nil) {
            [self appendQuoted:string string:pobox];
        }
        [string appendString:@";;"];
        if (street != nil) {
            [self appendQuoted:string string:street];
        }
        [string appendString:@";"];
        if (city != nil) {
            [self appendQuoted:string string:city];
        }
        [string appendString:@";"];
        if (state != nil) {
            [self appendQuoted:string string:state];
        }
        [string appendString:@";"];
        if (zip != nil) {
            [self appendQuoted:string string:zip];
        }
        [string appendString:@";"];
        if (country != nil) {
            [self appendQuoted:string string:country];
        }
        [string appendString:@"\n"];
    }
    
    [self appendStringProperty:string tnef:TNEF prodid:PR_CALLBACK_TELEPHONE_NUMBER format:@"TEL;X-EVOLUTION-CALLBACK:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_PRIMARY_TELEPHONE_NUMBER format:@"TEL;PREF:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_MOBILE_TELEPHONE_NUMBER format:@"TEL;CELL:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_RADIO_TELEPHONE_NUMBER format:@"TEL;X-EVOLUTION-RADIO:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_CAR_TELEPHONE_NUMBER format:@"TEL;CAR:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_OTHER_TELEPHONE_NUMBER format:@"TEL;VOICE:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_PAGER_TELEPHONE_NUMBER format:@"TEL;PAGER:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_TELEX_NUMBER format:@"TEL;X-EVOLUTION-TELEX:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_ISDN_NUMBER format:@"TEL;ISDN:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_HOME2_TELEPHONE_NUMBER format:@"TEL;HOME:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_TTYTDD_PHONE_NUMBER format:@"TEL;X-EVOLUTION-TTYTDD:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_HOME_TELEPHONE_NUMBER format:@"TEL;HOME;VOICE:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_ASSISTANT_TELEPHONE_NUMBER format:@"TEL;X-EVOLUTION-ASSISTANT:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_COMPANY_MAIN_PHONE_NUMBER format:@"TEL;WORK:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_BUSINESS_TELEPHONE_NUMBER format:@"TEL;WORK:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_BUSINESS2_TELEPHONE_NUMBER format:@"TEL;WORK;VOICE:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_PRIMARY_FAX_NUMBER format:@"TEL;PREF;FAX:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_BUSINESS_FAX_NUMBER format:@"TEL;WORK;FAX:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:PR_HOME_FAX_NUMBER format:@"TEL;HOME;FAX:%@\n"];
    
    // Email addresses
    if ((tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:0x8083]) == nil) {
        tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:0x8084];
    }
    if (tempString != nil) {
        if (tempString.length > 0)
            [string appendFormat:@"EMAIL:%@\n", tempString];
    }
    if ((tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:0x8093]) == nil) {
        tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:0x8094];
    }
    if (tempString != nil) {
        if (tempString.length > 0)
            [string appendFormat:@"EMAIL:%@\n", tempString];
    }
    if ((tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:0x80a3]) == nil) {
        tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:0x80a4];
    }
    if (tempString != nil) {
        if (tempString.length > 0)
            [string appendFormat:@"EMAIL:%@\n", tempString];
    }
    [self appendStringProperty:string tnef:TNEF prodid:PR_BUSINESS_HOME_PAGE format:@"URL:%@\n"];
    [self appendStringProperty:string tnef:TNEF prodid:0x80d8 format:@"FBURL:%@\n"];
    
    //Birthday
    if ((vl=MAPIFindProperty(&(TNEF.MapiProperties), PROP_TAG(PT_SYSTIME, PR_BIRTHDAY))) != MAPI_UNDEFINED) {
        [string appendString:@"BDAY:"];
        MAPISysTimetoDTR(vl->data, &thedate);
        [string appendFormat:@"%i-%02i-%02i\n", thedate.wYear, thedate.wMonth, thedate.wDay];
    }
    
    //Anniversary
    if ((vl=MAPIFindProperty(&(TNEF.MapiProperties), PROP_TAG(PT_SYSTIME, PR_WEDDING_ANNIVERSARY))) != MAPI_UNDEFINED) {
        [string appendString:@"X-EVOLUTION-ANNIVERSARY:"];
        MAPISysTimetoDTR(vl->data, &thedate);
        [string appendFormat:@"%i-%02i-%02i\n", thedate.wYear, thedate.wMonth, thedate.wDay];
    }
    
    // notes
    if ((tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:PR_PREVIEW]) != nil) {
        [string appendString:@"NOTES:"];
        [self appendQuoted:string string:tempString];
        [string appendString:@"\n"];
    }
    
    //
    if ((tempString = [self stringFromMapiProp:TNEF.MapiProperties propid:0x8062]) != nil) {
        [string appendFormat:@"IMPP:%@\n", tempString];
    }
    
    [string appendString:@"END:VCARD\n"];
    
    return string;
}

@end
