//
//  DetailView.m
//  iTNEFsEnough
//
//  Created by Josh Jacob on 11/15/15.
//  Copyright © 2015 Josh Jacob. All rights reserved.
//

#import "DetailView.h"

@interface DetailView ()

@property (nonatomic, weak) UIView *innerHeaderView;
@property (nonatomic, weak) IBOutlet UIView *headerView;

@end

@implementation DetailView

- (void)awakeFromNib
{
    [super awakeFromNib];
    
    [self createHeaderView];
    
    [self.webView setScalesPageToFit:YES];
    
    [self.webView.scrollView addObserver:self forKeyPath:@"contentOffset" options:NSKeyValueObservingOptionInitial context:nil];
    
    [self.webView.scrollView setShowsVerticalScrollIndicator:NO];
}

#pragma mark -
#pragma mark - NSKeyObserving

- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context
{
    // Update the frame of the header view so that it scrolls with the webview content
    CGRect newFrame = self.headerView.frame;
    newFrame.origin.y = -CGRectGetMinY([self.webView convertRect:self.innerHeaderView.frame toView:self.webView.scrollView]);
    [self.headerView setFrame:newFrame];
}

#pragma mark -
#pragma mark - Custom methods

- (void)createHeaderView
{
    UIView *headerView = [[UIView alloc] initWithFrame:CGRectMake(0, 0, CGRectGetWidth(self.frame), 83)];
    [headerView setBackgroundColor:[UIColor clearColor]];
    
    [self.webView.scrollView addSubview:headerView];
    [self setInnerHeaderView:headerView];
    
    [self addSubview:self.headerView];
}

#pragma mark -
#pragma mark - Custom layouting

- (void)layoutSubviews
{
    [super layoutSubviews];
    
    for (UIView *subview in self.webView.scrollView.subviews)
    {
        CGRect newFrame = subview.frame;
        if ([subview isEqual:self.innerHeaderView])
        {
            continue;
        }
        
        newFrame.origin.y = CGRectGetHeight(self.headerView.frame);
        [subview setFrame:newFrame];
    }
}

#pragma mark -
#pragma mark - Dealloc

- (void)dealloc
{
    [self.webView.scrollView removeObserver:self forKeyPath:@"contentOffset"];
}

@end
