//
//  DetailViewController.m
//  iTNEFsEnough
//
//  Created by Josh Jacob on 3/22/12.
//  Copyright (c) 2012 Josh Jacob. All rights reserved.
//

#import "DetailViewController.h"

#import "TNEFDataService.h"
#import "TNEFFileModel.h"

#import "FileCell.h"

#import "DetailView.h"

#import <QuickLook/QuickLook.h>

@interface DetailViewController ()
{
    BOOL isDirty;
}

- (void)configureView;

@end

@implementation DetailViewController

#pragma mark - Managing the detail item

- (void)setTnefFile:(TNEFFileModel *)newTnefFile
{
    if (_tnefFile != newTnefFile) {
        _tnefFile = newTnefFile;
        isDirty = YES;
    }
}

- (void)configureView
{
    if (_tnefFile == nil) {
        
        isDirty = YES;
        
        _embeddedFiles = @[];
        _docControllers = [[NSMutableArray alloc] init];
        
        // self.textViewFrom.text = @"";
        // self.textViewSubject.attributedText = [[NSMutableAttributedString alloc] initWithString:@""];
        self.detailView.labelFrom.text = @"";
        self.detailView.labelSubject.text = @"";
        [self.detailView.webView loadHTMLString:@"" baseURL:[NSURL URLWithString:@"/"]];
        
        self.detailView.hidden = YES;
        
        [_collectionViewFiles reloadData];
        
        return;
    }
    
    if (isDirty == NO) {
        return;
    }
    isDirty = NO;
    
    self.detailView.hidden = NO;
    
    _embeddedFiles = [TNEFDataService getFiles:_tnefFile];
    _docControllers = [[NSMutableArray alloc] initWithCapacity:_embeddedFiles.count];
    
    for (NSString *file in _embeddedFiles)
    {
        NSURL *url = [NSURL fileURLWithPath:file];
        UIDocumentInteractionController *docController = [UIDocumentInteractionController interactionControllerWithURL:url];
        docController.delegate = self;
        [_docControllers addObject:docController];
    }
    
    // Update the user interface for the detail item.
    
//    self.textViewFrom.text = [NSString stringWithFormat:@"From: %@", self.tnefFile.from];
//    if (self.tnefFile == nil || self.textViewFrom.text == nil || self.textViewFrom.text.length == 0) {
//        self.textViewFrom.text = @"From: <unknown sender>";
//    }
//    self.textViewFrom.font = [UIFont systemFontOfSize:13];
//    self.textViewFrom.textColor = [UIColor lightGrayColor];
    
    if (self.tnefFile.from == nil || self.tnefFile.from.length == 0) {
        self.detailView.labelFrom.text = @"<unknown sender>";
    } else {
        self.detailView.labelFrom.text = [NSString stringWithString:self.tnefFile.from];
    }
    
    if (self.tnefFile.subject == nil || self.tnefFile.subject.length == 0) {
        self.detailView.labelSubject.text = @"<no subject>";
    } else {
        self.detailView.labelSubject.text = [NSString stringWithString:self.tnefFile.subject];
    }
    
//    NSString *dateString = [NSDateFormatter localizedStringFromDate:self.tnefFile.dateSent dateStyle:NSDateFormatterLongStyle timeStyle:NSDateFormatterShortStyle];
//    if (self.tnefFile == nil || dateString == nil || dateString.length == 0) {
//        dateString = @"<unknown date>";
//    }
//    NSString *subjectDateString = [NSString stringWithFormat:@"%@\n%@", subjectString, dateString];
//    
//    NSMutableAttributedString *attrString = [[NSMutableAttributedString alloc] initWithString:subjectDateString];
//    NSRange subjectRange = NSMakeRange(0, subjectString.length);
//    [attrString addAttribute:NSFontAttributeName
//                       value:[UIFont boldSystemFontOfSize: 14]
//                       range:subjectRange];
//    NSRange dateRange = NSMakeRange(subjectString.length + 1, dateString.length);
//    [attrString addAttribute:NSForegroundColorAttributeName
//                       value:[UIColor lightGrayColor]
//                       range:dateRange];
//    [attrString addAttribute:NSFontAttributeName
//                       value:[UIFont systemFontOfSize: 13]
//                       range:dateRange];
//    self.textViewSubject.attributedText = attrString;

    // web view body
    if (self.tnefFile != nil) {
        if ([self.tnefFile.bodyIsHTML boolValue] == YES) {
            NSURL *fileUrl = [[[NSFileManager defaultManager] URLsForDirectory:NSDocumentDirectory inDomains:NSUserDomainMask] lastObject];
            fileUrl = [fileUrl URLByAppendingPathComponent:TNEF_DIR_FILES isDirectory:YES];
            fileUrl = [fileUrl URLByAppendingPathComponent:self.tnefFile.uuid isDirectory:YES];
            fileUrl = [fileUrl URLByAppendingPathComponent:@"body.html" isDirectory:NO];
            
            NSData *rtfData = [NSData dataWithContentsOfURL:fileUrl];
            [self.detailView.webView loadData:rtfData MIMEType:@"text/html" textEncodingName:@"utf-8" baseURL:[NSURL URLWithString:@"/"]];
        } else {
            NSURL *fileUrl = [[[NSFileManager defaultManager] URLsForDirectory:NSDocumentDirectory inDomains:NSUserDomainMask] lastObject];
            fileUrl = [fileUrl URLByAppendingPathComponent:TNEF_DIR_FILES isDirectory:YES];
            fileUrl = [fileUrl URLByAppendingPathComponent:self.tnefFile.uuid isDirectory:YES];
            fileUrl = [fileUrl URLByAppendingPathComponent:@"body.rtf" isDirectory:NO];
            
            NSData *rtfData = [NSData dataWithContentsOfURL:fileUrl];
            [self.detailView.webView loadData:rtfData MIMEType:@"text/rtf" textEncodingName:@"windows-1252" baseURL:[NSURL URLWithString:@"/"]];
        }
    }
    // [self.webViewBody sizeToFit];
    self.detailView.webView.scrollView.bounces = NO;
    self.detailView.webView.scrollView.scrollEnabled = YES;
    self.detailView.webView.scrollView.showsHorizontalScrollIndicator = NO;
    self.detailView.webView.scrollView.showsVerticalScrollIndicator = NO;
    
}

#pragma mark - View lifecycle

- (void)viewDidLoad
{
    [super viewDidLoad];
	// Do any additional setup after loading the view, typically from a nib.
    
    isDirty = YES;
    
    self.title = NSLocalizedString(@"Detail", @"Detail");
    
//    UIBarButtonItem *displayModeButtonItem = [self.splitViewController displayModeButtonItem];
//    
//    UINavigationItem *navItem = [[self.navigationController navigationBar] topItem];
//    
//    [navItem setLeftBarButtonItem:displayModeButtonItem animated:YES];
        
    self.navigationItem.leftBarButtonItem = self.splitViewController.displayModeButtonItem;
    self.navigationItem.leftItemsSupplementBackButton = YES;
}

- (void)viewWillAppear:(BOOL)animated
{
    [super viewWillAppear:animated];
    
    // Update the view.
    [self configureView];
}

-(UIInterfaceOrientationMask)supportedInterfaceOrientations
{
    return UIInterfaceOrientationMaskAll;
}

- (id)initWithNibName:(NSString *)nibNameOrNil bundle:(NSBundle *)nibBundleOrNil
{
    self = [super initWithNibName:nibNameOrNil bundle:nibBundleOrNil];
    if (self) {
        self.title = NSLocalizedString(@"Detail", @"Detail");
    }
    return self;
}


#pragma mark - UIWebViewDelegate

- (BOOL)webView:(UIWebView *)webView shouldStartLoadWithRequest:(NSURLRequest *)request navigationType:(UIWebViewNavigationType)navigationType
{
    if ((navigationType != UIWebViewNavigationTypeLinkClicked)
        && (navigationType != UIWebViewNavigationTypeFormSubmitted)
        && (navigationType != UIWebViewNavigationTypeBackForward)
        && (navigationType != UIWebViewNavigationTypeReload)
        && (navigationType != UIWebViewNavigationTypeFormResubmitted)) {
        return YES;
    }
    
    return NO;
}

#pragma mark - UICollectionViewDataSource

- (NSInteger)collectionView:(UICollectionView *)collectionView
     numberOfItemsInSection:(NSInteger)section
{
    return _embeddedFiles.count;
}

- (UICollectionViewCell *)collectionView:(UICollectionView *)collectionView
                           cellForItemAtIndexPath:(NSIndexPath *)indexPath
{
    FileCell *cell = (FileCell *) [collectionView dequeueReusableCellWithReuseIdentifier:@"FileCell"
                                                                           forIndexPath:indexPath];
    
    NSString *filePath = [_embeddedFiles objectAtIndex:[indexPath row]];
    UIDocumentInteractionController *docController = [_docControllers objectAtIndex:[indexPath row]];
    cell.fileLabel.text = [filePath lastPathComponent];
    NSInteger iconCount = [docController.icons count];
    if (iconCount > 0)
    {
        cell.fileImage.image = [docController.icons objectAtIndex:iconCount - 1];
    }
    
    cell.contentView.gestureRecognizers = docController.gestureRecognizers;
    
    return cell;
}

#pragma mark - UICollectionViewDelegate

#pragma mark - UIDocumentInteractionControllerDelegate

- (UIViewController *)documentInteractionControllerViewControllerForPreview:(UIDocumentInteractionController *)controller
{
    if (self.navigationController != nil)
        return self.navigationController;
    
    return self.splitViewController;
}

@end
